
/* $Header: /usr/src/local/etc/tftpd/RCS/udp_syslog.c,v 1.4 1995/02/20 21:14:38 smb Exp $ */

/* 
 * udp_syslog.c - adaptation of syslog library routines to use UDP
 * 
 * Modified by:	Scott M. Ballew
 * 		Purdue University Computing Center
 * 		Purdue University
 * Date:	Tue Mar 15 10:06:50 1994
 *
 * Modifications Copyright 1994 Purdue Research Foundation,
 *	West Lafayette, Indiana 47907.  All rights reserved.
 *
 * Written by Scott M. Ballew
 *
 * The modifications in this software are not subject to any license of
 * the American Telephone and Telegraph Company or the Regents of the
 * University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

/*
 * NOTE: This software is derived from the NET2 syslog.c library routines
 * which are covered by the following copyright notice and use provisions.
 */

/*
 *
 * Copyright (c) 1983, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifdef ACCESS_CONTROL

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)syslog.c	5.36 (Berkeley) 10/4/92";
#endif /* LIBC_SCCS and not lint */

/*
 * $Log: udp_syslog.c,v $
 * Revision 1.4  1995/02/20  21:14:38  smb
 * Fix a botch in the earlier changes.
 *
 * Revision 1.3  1995/02/20  21:11:14  smb
 * Cleaned up Purdue Copyright information.
 *
 * Revision 1.2  1994/03/16  21:55:39  smb
 * Fixed to log to primary interface address so hostname is better.
 * Still need to get timezone processing working.
 *
 * Revision 1.1  1994/03/15  16:02:56  smb
 * Adapted to use UDP rather than Unix socket to communicate with
 * syslogd.  This allow syslog to work after a chroot().  These routines
 * are based on the NET2 syslog routines, with a small piece of a prior
 * BSD release to handle logging to the console.
 *
 */

#ifndef lint
static char rcs_ident[] = "$Id: udp_syslog.c,v 1.4 1995/02/20 21:14:38 smb Exp $";
#endif lint

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1983,1988 Regents of the University of California.\n\
 All rights reserved.\n\
 Modifications Copyright 1994 Purdue Research Foundation.\n\
 All rights reserved.\n";
#endif lint

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/sockio.h>

#include <netdb.h>
#include <fcntl.h>
#include <syslog.h>
#include <varargs.h>
#include <signal.h>

#ifndef NULL
#define NULL	0
#endif

#ifndef LOG_PRI
#define LOG_PRI(p)	((p) & LOG_PRIMASK)
#endif

static int	LogFile = -1;		/* fd for log */
static int	connected;		/* have done connect */
static int	LogStat = 0;		/* status bits, set by openlog() */
static char 	*LogTag = "syslog";	/* string to tag the entry with */
static int	LogFacility = LOG_USER;	/* default facility code */
static int	LogMask = 0xff;		/* mask of priorities to be logged */

static char 	*ctty = "/dev/console";

void UDP_vsyslog();
void UDP_openlog();

extern int errno;

/*
 * syslog, vsyslog --
 *	print message on log file; output is intended for syslogd(8).
 */
void UDP_syslog(pri, fmt, va_alist)
     int pri;
     char *fmt;
     va_dcl
{
    va_list ap;

    va_start(ap);
    UDP_vsyslog(pri, fmt, ap);
    va_end(ap);
}

void UDP_vsyslog(pri, fmt, ap)
     int pri;
     register char *fmt;
     va_list ap;
{
    register int cnt;
    register char *p;
    time_t now;
    int fd, saved_errno, pid;
    char tbuf[2048], fmt_cpy[1024];
    
#define	INTERNALLOG	LOG_ERR|LOG_CONS|LOG_PID
    /* Check for invalid bits. */
    if (pri & ~(LOG_PRIMASK|LOG_FACMASK)) {
	syslog(INTERNALLOG,
	       "syslog: unknown facility/priority: %x", pri);
	pri &= LOG_PRIMASK|LOG_FACMASK;
    }
    
    /* Check priority against setlogmask values. */
    if (!LOG_MASK(LOG_PRI(pri)) & LogMask)
	return;
    
    saved_errno = errno;
    
    /* set default facility if none specified */
    if ((pri & LOG_FACMASK) == 0)
	pri |= LogFacility;
    
    /* build the message */
    (void)time(&now);
    (void)sprintf(tbuf, "<%d>%.15s ", pri, ctime(&now) + 4);
    for (p = tbuf; *p; ++p);
    if (LogTag) {
	(void)strcpy(p, LogTag);
	for (; *p; ++p);
    }
    if (LogStat & LOG_PID) {
	(void)sprintf(p, "[%d]", getpid());
	for (; *p; ++p);
    }
    if (LogTag) {
	*p++ = ':';
	*p++ = ' ';
    }
    
    /* substitute error message for %m */
    {
	register char ch, *t1, *t2;
	char *strerror();
	
	for (t1 = fmt_cpy; ch = *fmt; ++fmt)
	    if (ch == '%' && fmt[1] == 'm') {
		++fmt;
		for (t2 = strerror(saved_errno);
		     *t1 = *t2++; ++t1);
	    }
	    else
		*t1++ = ch;
	*t1 = '\0';
    }
    
    (void)vsprintf(p, fmt_cpy, ap);
    
    cnt = strlen(tbuf);
    
    /* get connected, output the message to the local logger */
    if (!connected)
	UDP_openlog(LogTag, LogStat | LOG_NDELAY, 0);
    if (send(LogFile, tbuf, cnt, 0) >= 0)
	return;

    /* see if should attempt the console */
    if (!(LogStat&LOG_CONS))
	return;
    
    /* output the message to the console */
    pid = vfork();
    if (pid == -1)
	return;
    if (pid == 0) {
	int fd;
	
	(void)signal(SIGALRM, SIG_DFL);
	(void)sigsetmask(sigblock(0) & ~sigmask(SIGALRM));
	(void)alarm(5);
	fd = open(ctty, O_WRONLY);
	(void)alarm(0);
	(void)strcat(p, "\r");
	p = (char *) index(tbuf, '>') + 1;
	(void)write(fd, p, cnt + 1 - (p - tbuf));
	(void)close(fd);
	_exit(0);
    }
    if (!(LogStat & LOG_NOWAIT))
	while ((cnt = wait((int *)0)) > 0 && cnt != pid)
	    ;
}

#ifndef SYSLOG_PORT
#define SYSLOG_PORT	514
#endif

static struct sockaddr_in SyslogAddr;	/* AF_INET address of local logger */

static char buffer[3*sizeof(struct ifreq)];
     
void UDP_openlog(ident, logstat, logfac)
     char *ident;
     int logstat, logfac;
{
    struct ifconf ifc;
    register struct sockaddr_in *sin;
    register int i;

    if (ident != NULL)
	LogTag = ident;
    LogStat = logstat;
    if ((logfac != 0) && ((logfac & ~LOG_FACMASK) == 0))
	LogFacility = logfac;

    if (LogFile == -1) {
	struct servent *sp;

	SyslogAddr.sin_family = AF_INET;
	sp = getservbyname("syslog", "udp");
	if (sp != NULL) {
	    SyslogAddr.sin_port = sp->s_port;
	}
	else {
	    SyslogAddr.sin_port = htons(SYSLOG_PORT);
	}
	if (LogStat & LOG_NDELAY) {
	    if ((LogFile = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
		return;
	    (void) fcntl(LogFile, F_SETFD, 1);
	}
	ifc.ifc_len = sizeof(buffer);
	ifc.ifc_buf = buffer;
	if (ioctl(LogFile, SIOCGIFCONF, &ifc) < 0) {
	    SyslogAddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	}
	else {
	    ifc.ifc_len /= sizeof(struct ifreq);
	    for(i=0; i<ifc.ifc_len; i++) {
		sin = (struct sockaddr_in *) &ifc.ifc_req[i].ifr_addr;
		if ((sin->sin_addr.s_addr) &&
		    (sin->sin_addr.s_addr != INADDR_LOOPBACK)) {
		    break;
		}
	    }
	    if (i < ifc.ifc_len)
		SyslogAddr.sin_addr.s_addr = sin->sin_addr.s_addr;
	    else
		SyslogAddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	}
    }
    if ((LogFile != -1) && (!connected))
	if (connect(LogFile, &SyslogAddr, sizeof(SyslogAddr)) == -1) {
	    (void) close(LogFile);
	    LogFile = -1;
	}
    else
	connected = 1;
}

void UDP_closelog()
{
    (void)close(LogFile);
    LogFile = -1;
    connected = 0;
}

/* setlogmask -- set the log mask level */
int UDP_setlogmask(pmask)
     int pmask;
{
    int omask;
    
    omask = LogMask;
    if (pmask != 0)
	LogMask = pmask;
	return (omask);
}

#endif /* ACCESS_CONTROL */
